#!/usr/bin/perl -w

# reads and summarizes the output of dumpe2fs
# expects non-localized dumpe2fs output, so call dumpe2fs with LANG=C

$n_groups = 0;    # number of read block groups
$color = 1;       # if 1, block groups with critical values are marked in bold

while (<STDIN>)
{
  chomp;
  if (!defined($all_inodes)) { /Inode count\D+([0-9]+)/ and $all_inodes = $1; }
  if (!defined($all_blocks)) { /Block count\D+([0-9]+)/ and $all_blocks = $1; }
  if (!defined($free_blocks)) { /Free blocks\D+([0-9]+)/ and $free_blocks = $1; }
  if (!defined($free_inodes)) { /Free inodes\D+([0-9]+)/ and $free_inodes = $1; }
  /Blocks per group\D+([0-9]+)/ and $n_blocks = $1;    # number of blocks per group
  /Inodes per group\D+([0-9]+)/ and $n_inodes = $1;    # number of inodes per group
  /^Group ([0-9]+)/ and $group=$1;      # block group number, start of group statistics
  defined($group) or next;              # skip header lines

# extract statistics for one block group
  if (/([0-9]+) free blocks, ([0-9]+) free inodes, ([0-9]+) directories/)
  {                                     # basic block group statistics
    $data[$group]{'blocks'} = $1;
    $data[$group]{'inodes'} = $2;
    $data[$group]{'dirs'} = $3;
    $n_groups++;
  }

  if (/Free blocks/)                   # list of free blocks: chunks (nnn-nnn) or single blocks (nnn)
  {
    if (@matches = /([0-9]+-[0-9]+)/g)
    {
      $data[$group]{'chunks'} = scalar @matches;
    }
    else { $data[$group]{'chunks'} = 0; }
    if (@matches = / ([0-9]+),/g)      # ' ' and ',' distinguish from start or end block of chunk
    {
      $data[$group]{'single'} = scalar @matches;
    }
    else { $data[$group]{'single'} = 0; }
    / [0-9]+$/ and $data[$group]{'single'}++;   # single block at EOL
  }
}

# display statistics of block groups
# logic for "colorizing" output comes from summary statistics
print "$all_blocks blocks ($free_blocks free), $all_inodes ($free_inodes free)\n";
print "$n_groups block groups with $n_blocks blocks and $n_inodes inodes each:\n";
for ($i=0; $i<$n_groups; $i++)
{
  printf ("Block group %3d: ", $i);
  $data[$i]{'inodes'} < $n_inodes/16 and ($color and print "\033[1m");
  print "$data[$i]{'inodes'} free inodes";
  $data[$i]{'inodes'} < $n_inodes/16 and ($color and print "\033[m");
  print ", $data[$i]{'dirs'} directories;\n  ";
  ($data[$i]{'blocks'} > 4 and $data[$i]{'chunks'}+$data[$i]{'single'} > $data[$i]{'blocks'}/4)
    and ($color and print "\033[1m");
  print "$data[$i]{'blocks'} free blocks in $data[$i]{'chunks'} chunks, ";
  print "$data[$i]{'single'} single blocks";
  ($data[$i]{'blocks'} > 4 and $data[$i]{'chunks'}+$data[$i]{'single'} > $data[$i]{'blocks'}/4)
    and ($color and print "\033[m");
  if ($data[$i]{'chunks'}+$data[$i]{'single'} > 0)
  {
    $avg = $data[$i]{'blocks'}*1.0 / ($data[$i]{'chunks'}+$data[$i]{'single'});
    printf (" (avg. chunk size: %3.1f)", $avg);
  }
  print "\n";
}

# summary of block groups with critical values
print "\nSummary: $all_blocks blocks ($free_blocks free), $all_inodes inodes ($free_inodes free)\n";
print "$n_groups block groups with $n_blocks blocks and $n_inodes inodes each:\n\n";
$n16 = 0;
$n8 = 0;
$n4 = 0;
$n2 = 0;
$n = 0;
$n0 = 0;
for ($i=0; $i<$n_groups; $i++)
{
  if ($data[$i]{'inodes'} == 0) { $n0++; }                          # no free inodes
  elsif ($data[$i]{'inodes'} < $n_inodes/16) { $n16++; }  # only 1/16 if inodes free: worst
  elsif ($data[$i]{'inodes'} < $n_inodes/8) { $n8++; }    # < 1/8: worse
  elsif ($data[$i]{'inodes'} < $n_inodes/4) { $n4++; }    # < 1/4: bad
  elsif ($data[$i]{'inodes'} < $n_inodes/2) { $n2++; }    # < 1/2: ok
  else { $n++; }                                          # > 1/2: fine, fine
}
print "$n block groups with many (> 50%) free inodes,\n";
print "$n2 block groups with enough (< 1/2) free inodes,\n";
print "$n4 block groups with few (< 1/4) free inodes,\n";
print "$n8 block groups with very few (< 1/8) free inodes,\n";
print "$n16 block groups with *really* few (< 1/16) free inodes,\n";
print "$n0 block groups with no free inodes.\n";

$n64 = 0;
$n32 = 0;
$n16 = 0;
$n8 = 0;
$n4 = 0;
$n = 0;
$n0 = 0;
for ($i=0; $i<$n_groups; $i++)
{
  if ($data[$i]{'blocks'} == 0) { $n0++; }     # no free blocks
  elsif ($data[$i]{'blocks'} > 4 and $data[$i]{'chunks'}+$data[$i]{'single'} > $data[$i]{'blocks'}/4)     # worst:
   { $n4++; }          # more than 4 free blocks that are fragmented across more than (free blocks/4) chunks
  elsif ($data[$i]{'blocks'} > 8 and $data[$i]{'chunks'}+$data[$i]{'single'} > $data[$i]{'blocks'}/8)  # worse
    { $n8++; }
  elsif ($data[$i]{'blocks'} > 16 and $data[$i]{'chunks'}+$data[$i]{'single'} > $data[$i]{'blocks'}/16)  # bad
   { $n16++; }
  elsif ($data[$i]{'blocks'} > 32 and $data[$i]{'chunks'}+$data[$i]{'single'} > $data[$i]{'blocks'}/32)  # ok?
   { $n32++; }
  elsif ($data[$i]{'blocks'} > 64 and $data[$i]{'chunks'}+$data[$i]{'single'} > $data[$i]{'blocks'}/64)  # ok?
   { $n64++; }
  else { $n++; }
}
print "\n$n block groups with average chunk size > 64 blocks,\n";
print "$n64 block groups with average chunk size < 64 blocks,\n";
print "$n32 block groups with average chunk size < 32 blocks,\n";
print "$n16 block groups fragmented with average chunk size < 16 blocks,\n";
print "$n8 block groups badly fragmented with average chunk size < 8 blocks,\n";
print "$n4 block groups heavily fragmented with average chunk size < 4 blocks,\n";
print "$n0 block groups without free blocks.\n";
