/* scans files on an ext3 fs for fragmentation */
/* compile with: gcc -O2 -lext2fs -o ext2_frag ext2_frag.c */

/* scans an inode or all inodes of a file system for fragmentation */

/* 0.1: seems to work */
/* 0.2: switched from ext2_get_blocks and manual evaluation of indirect */
/*      blocks to ext2fs_block_iterate() */
/* 0.3: inode data are not stored, but printed as evaluated. Saves hundreds */
/*      of MBytes RAM on large file systems */
/* 0.4: removed a big memory hole; block function simplified */
/* 0.6: added fragmentation index: number of block jumps > 1 block (number */
/*      of fragments - 1) divided by possible jumps (number of blocks - 1) */
/*      across all files; options -s, -f; formatting of output */



/* BUGS */
/* probably ... */
/* on one file > 1 GByte ext2_frag finds less fragments than filefrag, but */
/* on several files with some hundred MByte (spanning several block groups), */
/* they agree. Don't understand what's wrong. */

/* (C) Oliver Diedrich 2007 */
/* this file may be redistributed under the terms of the */
/* GNU Public License (GPL), see www.gnu.org */

#define VERSION "0.6"

#define _LARGEFILE64_SOURCE
#define _FILE_OFFSET_BITS 64

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <ext2fs/ext2fs.h>
#include <time.h>
#include <fcntl.h>
#include <unistd.h>


struct statistics            /* summary statistics for different file sizes */
{
  unsigned long n;           /* number of files in this class */
  unsigned long n_frag;      /* number of fragmented files */
  unsigned long frag;        /* number of fragments */
  unsigned long n_blocks;    /* number of blocks */
} stat [6];


struct expected              /* expected values for the recent block */
{                            /* for ext2fs_block_iterate() */
  unsigned long nr;          /* block number */
  unsigned long count;       /* logical block number (0, 1, ...) */
  unsigned long n_frag;      /* number of fragments in file processed */
  int verbose;               /* -1: summary only; 1: verbose, 2: debugging info */
  unsigned long entries;     /* # block numbers in an indirect block */
};



void usage(char *name)
/* prints usage information */
{
  printf("Usage: %s device [-s|-f|-v|-d] [-i Inode]\n", name);
  printf("-s: summary statistics only (default: statistics for each file)\n");
  printf("-f: print only fragmented files (default: statistics for each file)\n");
  printf("-v: be verbose when checking\n");
  printf("-d: even more verbose (for debugging only)\n");
  printf("-i Inode: check only this inode (default: check whole file system)\n");
  _exit(1);
}



ext2_filsys open_fs(char *dev_name)
/* opens the file system dev_name */
/* NULL: error, else: file system handle */
{
  int err;
  ext2_filsys fs;
  char buf[PATH_MAX+30];

  err = ext2fs_open(dev_name,   /* /dev/... */
                    0,          /* read_only, use superblock feature sets */
                    0,          /* use primary superblock */
                    0,          /* use block size from superblock */
                    unix_io_manager,    /* unix_io IO manager */
                    &fs);       /* filesystem handle */
  if (err)
  {
    sprintf(buf, "while opening %s", dev_name);
    com_err(dev_name, err, buf);
    return NULL;
  }

  printf("%s opened for %s access\n", dev_name,
         fs->flags & EXT2_FLAG_RW ? "read-write" : "read-only");
  if (fs->flags &&  EXT2_FLAG_DIRTY)
  {
    printf("%s was not cleanly unmounted, there might be inconsistencies.\n",
           dev_name);
  }
  return fs;
}   /* open_fs() */



void show_fs_info(ext2_filsys fs)
/* print some superblock content of file system fs */
{
  printf("%d inodes (%d free)\n", fs->super->s_inodes_count,
         fs->super->s_free_inodes_count);
  printf("%d blocks of %d bytes (%d free)\n", fs->super->s_blocks_count,
         1024 << fs->super->s_log_block_size, fs->super->s_free_blocks_count);
  printf("last mounted on %s", ctime( (time_t *) &(fs->super->s_mtime)) );
}   /* show_fs_info() */



int count_frags(ext2_filsys fs,
                blk_t *blocknr,       /* number of recent block */
                int blockcnt,         /* logical count of recent block */
                struct expected *exp) /* contains expected values of blocknr */ 
                                      /* and blockcnt, number of fragments, */
                                      /* and the number of block numbers in */
                                      /* an indirect block */
/* called by ext2fs_block_iterate() in process_blocks() */
/* determines fragmentation and counts number of fragments */
/* returns always 0 */

/* Perfect layout of a file: */
/* B B B B B B B B B B B B I B B .... B B D I B B ... B B I B B ... */
/* 12 (direct) data blocks [B], indirect block [I], data blocks, */
/* double indirect block [D], indirect block, data blocks, ... */

{
  if ( (blockcnt<0) && (2==exp->verbose) )    /* print metadata blocks for */
  {                                           /* debugging */
    printf("---------------------------\n");
    printf("expected: %6lu: %8lu\n", exp->count, exp->nr);
    if (BLOCK_COUNT_IND == blockcnt)
     printf(" got IND: %6d: %8lu\n", blockcnt, (unsigned long) *blocknr);
    else if (BLOCK_COUNT_DIND == blockcnt)
     printf("got DIND: %6d: %8lu\n", blockcnt, (unsigned long) *blocknr);
    else if (BLOCK_COUNT_TIND == blockcnt)
     printf("got TIND: %6d: %8lu\n", blockcnt, (unsigned long) *blocknr);
    else if (BLOCK_COUNT_TRANSLATOR == blockcnt)
     printf(" got TRA: %6d: %8lu\n", blockcnt, (unsigned long) *blocknr);
    else 
     printf(" got ???: %6d: %8lu\n", blockcnt, (unsigned long) *blocknr);
  }

  if (*blocknr != exp->nr) /* not the expected block number: fragmentation */
  {
    exp->n_frag++;
    if (exp->verbose>0)    /* print it for information */
    {
      {
        printf("---------------------------\n");
        printf("expected: %6lu: %8lu\n", exp->count, exp->nr);
        if (BLOCK_COUNT_IND == blockcnt)
         printf(" got IND: %6d: %8lu", blockcnt, (unsigned long) *blocknr);
        else if (BLOCK_COUNT_DIND == blockcnt)
         printf("got DIND: %6d: %8lu", blockcnt, (unsigned long) *blocknr);
        else if (BLOCK_COUNT_TIND == blockcnt)
         printf("got TIND: %6d: %8lu", blockcnt, (unsigned long) *blocknr);
        else if (BLOCK_COUNT_TRANSLATOR == blockcnt)
         printf(" got TRA: %6d: %8lu", blockcnt, (unsigned long) *blocknr);
        else 
         printf("     got: %6d: %8lu", blockcnt, (unsigned long) *blocknr);
        printf("    (%lu fragments)\n", exp->n_frag);
      }
    }
  }
  if (blockcnt >= 0)                        /* not a meta block */
   exp->count = (unsigned long) blockcnt+1; /* logical block count and block */
  exp->nr = (unsigned long) (*blocknr)+1;   /* number should be consecutive */
  return 0;
} /* count_frags() */



int process_blocks(ext2_filsys fs, ext2_ino_t inode, int verbose)
/* iterates over the blocks of inode, calling block_number() for each block */
/* 0: error, else: number of fragments in inode */
{
  char *block_buf;    /* for ext2_block_iterate(); must be at least three */
  errcode_t err;                                  /* times the block size */
  struct expected exp;
  
  exp.count = 0;            /* the expectd next block count */
  exp.nr = 0;               /* the expected block numer */
  exp.n_frag = 0;           /* counts the number of fragments */
  exp.verbose = verbose;    /* print debug information */
  exp.entries = (1024 << fs->super->s_log_block_size) / sizeof(blk_t);
                            /* number of block numbers in an indirect block */
  block_buf = calloc(4, 1024 << fs->super->s_log_block_size);
  err = ext2fs_block_iterate(fs,           /* file system */
                             inode,        /* inode number */
                             0,            /* no flags */
                             block_buf,    /* buffer */
                             count_frags,  /* function called for each block */
                             &exp);        /* data for count_frags() */
  free(block_buf);
  if (err) return 0;
  else return exp.n_frag;
} /* process_blocks() */



int scan_inodes(ext2_filsys fs, int verbose)
/* reads and analyzes all inodes in fs */
/* 0:error, 1: number of inodes read */
{
  ext2_inode_scan scan;
  errcode_t err;
  ext2_ino_t inode;
  struct ext2_inode inode_struct;
  unsigned long n = 0;                /* number of valid inodes */
  int bsize;                          /* block size */
  int st;                             /* size class of file */
  char buf[255];
  unsigned long n_block;              /* number of blocks in recent inode */
  unsigned long frag;                 /* number of fragments in recent inode */

  err = ext2fs_open_inode_scan(fs,    /* file system */
                               0,     /* use suitable value for block buffer */
                               &scan);/* inode iteration variable */
  if (err)
  {
    com_err("scan_inodes()", err, "while initializing inode scan");
    return 0;
  }
  else printf("\nScanning all inodes (this can take a while ...)\n");
  if (verbose > -2)
  {
    printf("type    Inode   blocks  fragments  blocks/fragment  fragm. index\n");
    printf("----------------------------------------------------------------\n");
  }

  bsize = 1024 << fs->super->s_log_block_size;

  do
  {
    err = ext2fs_get_next_inode(scan, &inode, &inode_struct);
    if (err)              /* only EXT2_ET_BAD_BLOCK_IN_INODE_TABLE possible? */
    {
      sprintf(buf, "while getting inode %lu", n-1);
      com_err("scan_inodes()", err, buf);
      continue;
    }

/* some inodes aren't checked: */
    if (!inode) continue;    /* no valid inode */
    if (inode < (fs->super->s_first_ino))    /* reserved inode (journal etc) */
    {
      if (verbose>0) printf("Inode %lu: reserved\n", (unsigned long) inode);
      continue;
    }
    if (LINUX_S_ISCHR(inode_struct.i_mode)) continue;   /* character device */
    if (LINUX_S_ISBLK(inode_struct.i_mode)) continue;   /* block device */
    if (LINUX_S_ISFIFO(inode_struct.i_mode)) continue;  /* pipe */
    if (LINUX_S_ISSOCK(inode_struct.i_mode)) continue;  /* socket */

/* number of blocks of this file. The field i_blocks in the inode structure */
/* contains the number of 512 Byte sectors, not ext2 blocks! */
    n_block = inode_struct.i_blocks / (2 << fs->super->s_log_block_size);
    if (!n_block) continue;

    if (n_block == 1) frag = 1;               /* only 1 block: can't fragment */
    else
    {
      if (verbose>0) printf("* Inode %5lu\n", (unsigned long) inode);
      frag = process_blocks(fs, inode, verbose);
      if (!frag) continue;   /* process_blocks() returned an error */
    }
    n++;

/* file size classes: */
    if (n_block == 1) st = 0;
    else if (n_block <= EXT2_NDIR_BLOCKS) st = 1;
    else if (n_block <= bsize / sizeof(blk_t) / 2 + EXT2_NDIR_BLOCKS) st = 2;
    else if (n_block <= bsize / sizeof(blk_t) + EXT2_NDIR_BLOCKS) st = 3;
    else if (n_block <= bsize / sizeof(blk_t) * 4 + EXT2_NDIR_BLOCKS) st = 4;
    else st = 5;
    stat[st].n++;
    stat[st].frag += frag;
    stat[st].n_blocks += n_block;

    if (verbose > -2)
    {
      if (frag > 1)
      {
        if (LINUX_S_ISREG(inode_struct.i_mode)) printf("(-) ");   /* file type */
        else if (LINUX_S_ISLNK(inode_struct.i_mode)) printf("(l) ");
        else if (LINUX_S_ISDIR(inode_struct.i_mode)) printf("(d) ");
        else printf("(?) ");
        printf("%8lu: %8lu  %8lu  %15.1f  %11.3f%%\n",
               (unsigned long) inode, n_block, frag, n_block*1.0 / frag, 
               (frag-1)*100.0 / (n_block-1));
      }
      else
      {
        if (verbose > -1)
        {
          if (LINUX_S_ISREG(inode_struct.i_mode)) printf("(-) ");   /* file type */
          else if (LINUX_S_ISLNK(inode_struct.i_mode)) printf("(l) ");
          else if (LINUX_S_ISDIR(inode_struct.i_mode)) printf("(d) ");
          else printf("(?) ");
          printf("%8lu: %8lu   unfragmented\n", (long) inode, n_block);
        }
      }
    }
    if (frag>1) stat[st].n_frag++;

  } while (inode);

  ext2fs_close_inode_scan(scan);
  return n;
}   /* scan_inodes() */



int main(int argc, char *argv[])
{
  char device[PATH_MAX];      /* file system path to check for fragmentation */
  ext2_filsys fs;             /* ext2 file system, for low level access */
  int i;
  char *c;
  int verbose = 0;            /* -v / -d */
  unsigned long inode = 0;    /* inode number (-i) */
  errcode_t err;
  struct ext2_inode inode_struct;
  unsigned long n_files;      /* number of tested files */
  unsigned long n_frag;       /* number of fragmented files */
  unsigned long frag;         /* number of fragments */
  unsigned long n_blocks = 0; /* number of blocks */


  printf("%s %s\n", argv[0], VERSION);
  if (argc<2)
  {
    usage(argv[0]);
  }
  for (i=1; i<argc; i++)
  {
    c = argv[i];
    if ('-' == *c)
    {
      c++;
      switch (*c)
      {
        case 's' : verbose = -2; break;
        case 'f' : verbose = -1; break;
        case 'v' : verbose = 1; break;
        case 'd' : verbose = 2 ; break;
        case 'i' : if (++i < argc) inode = atol(argv[i]);
                   else usage(argv[0]);
                   if (!inode) usage(argv[0]);
                   break;
        default :  usage(argv[0]);
      }
    }
    else sprintf(device, argv[i]);
  }

  if (!(fs = open_fs(device))) return 1;             /* open file system ... */
  show_fs_info(fs);               /* ,,, and show some superblock statistics */

/* check only one inode */
  if (inode)
  {
    err = ext2fs_read_inode(fs, (ext2_ino_t) inode, &inode_struct);
    if (err)
    {
      com_err("ext2fs_read_inode()", err, "while reading inode");
      ext2fs_close(fs);
      return 1;
    }
    frag = process_blocks(fs, inode, verbose);

    if (verbose>0) printf("\n");
    if (frag > 1)
     printf("Inode %lu: %6lu blocks in %4lu fragments\n", inode,
      (unsigned long) inode_struct.i_blocks / (2<<fs->super->s_log_block_size),
      frag);
    else printf("Inode %lu: %6lu blocks, unfragmented\n", inode,
      (unsigned long) inode_struct.i_blocks / (2<<fs->super->s_log_block_size));
    if (ext2fs_close(fs))
    {
      fprintf(stderr, " %s: error while closing\n", device);
      return 1;
    }
    return 0;
  }

/* check whole file system */
  for (i=0; i<6; i++)    /* intialize global variable stat[] which is filled */
  {                      /* by scan_inodes() */
    stat[i].n = 0;
    stat[i].n_frag = 0;
    stat[i].frag = 0;
    stat[i].n_blocks = 0;
  }
    
  inode = scan_inodes(fs, verbose);

/* statistics for the file size classes */
  printf("\nSummary for %s (block size %d bytes):\n", device, 
         1024 << fs->super->s_log_block_size);
  printf("number of blocks    files  fragmented  percent  fragments/file\n");
  printf("--------------------------------------------------------------\n");
  printf("              1 %8lu           0      0.00         1.00\n", 
         stat[0].n);
  printf("        <= %4u %8lu    %8lu  %8.2f  %11.2f\n",
         EXT2_NDIR_BLOCKS, 
         stat[1].n, stat[1].n_frag, 
         stat[1].n > 0 ? stat[1].n_frag*100.0 / stat[1].n : 0,
         stat[1].n > 0 ? stat[1].frag*1.0 / stat[1].n : 0);
  printf("        <= %4u %8lu    %8lu  %8.2f  %11.2f\n", 
   (1024 << fs->super->s_log_block_size) / sizeof(blk_t) / 2 + EXT2_NDIR_BLOCKS,
   stat[2].n, stat[2].n_frag, 
   stat[2].n > 0 ? stat[2].n_frag*100.0 / stat[2].n : 0,
   stat[2].n > 0 ? stat[2].frag*1.0 / stat[2].n : 0);
  printf("        <= %4u %8lu    %8lu  %8.2f  %11.2f\n", 
   (1024 << fs->super->s_log_block_size) / sizeof(blk_t) + EXT2_NDIR_BLOCKS,
   stat[3].n, stat[3].n_frag,
   stat[3].n > 0 ? stat[3].n_frag*100.0 / stat[3].n : 0,
   stat[3].n > 0 ? stat[3].frag*1.0 / stat[3].n : 0);
  printf("        <= %4u %8lu    %8lu  %8.2f  %11.2f\n",
   (1024 << fs->super->s_log_block_size) / sizeof(blk_t) * 4 + EXT2_NDIR_BLOCKS,
   stat[4].n, stat[4].n_frag,
   stat[4].n > 0 ? stat[4].n_frag*100.0 / stat[4].n : 0,
   stat[4].n > 0 ? stat[4].frag*1.0 / stat[4].n : 0);
  printf("         > %4u %8lu    %8lu  %8.2f  %11.2f\n",
   (1024 << fs->super->s_log_block_size) / sizeof(blk_t) * 4 + EXT2_NDIR_BLOCKS,
   stat[5].n, stat[5].n_frag, 
   stat[5].n > 0 ? stat[5].n_frag*100.0 / stat[5].n : 0,
   stat[5].n > 0 ? stat[5].frag*1.0 / stat[5].n : 0);
  printf("--------------------------------------------------------------\n");

/* compute summary statistics for file system */
  n_files = 0;
  n_frag = 0;
  frag = 0;
  for (i=0; i<6; i++)
  {
    n_files += stat[i].n;
    n_frag += stat[i].n_frag;
    frag += stat[i].frag;
    n_blocks += stat[i].n_blocks;
  }
  printf("all files       %8lu    %8lu  %8.2f  %11.2f\n",
         n_files, n_frag, 
         n_files > 0 ? n_frag*100.0 / n_files : 0,
         n_files > 0 ? frag*1.0 / n_files : 0);
  n_files = 0;
  n_frag = 0;
  frag = 0;
  for (i=1; i<6; i++)
  {
    n_files += stat[i].n;
    n_frag += stat[i].n_frag;
    frag += stat[i].frag;
  }
  printf("files > 1 block %8lu    %8lu  %8.2f  %11.2f\n",
         n_files, n_frag, 
         n_files > 0 ? n_frag*100.0 / n_files : 0,
         n_files > 0 ? frag*1.0 / n_files : 0);
/* number of jumps between blocks, divided by the number of possible jumps */
  printf("Fragmentation index: %6.3f percent.\n", 
         (frag - n_files) * 100.0 / (n_blocks - n_files));

  if (ext2fs_close(fs))
  {
    fprintf(stderr, " %s: error while closing\n", device);
    return 1;
  }
  return 0;
}
